from ..types.location import Location

#
# Constants
#

abovebar = Location('AboveBar')
absolute = Location('Absolute')
belowbar = Location('BelowBar')
bottom = Location('Bottom')
top = Location('Top')
