from ..types.currency import Currency

#
# Constants
#

AUD = Currency('AUD')
BTC = Currency('BTC')
CAD = Currency('CAD')
CHF = Currency('CHF')
ETH = Currency('ETH')
EUR = Currency('EUR')
GBP = Currency('GBP')
HKD = Currency('HKD')
JPY = Currency('JPY')
KRW = Currency('KRW')
MYR = Currency('MYR')
NOK = Currency('NOK')
NONE = Currency('NONE')
NZD = Currency('NZD')
RUB = Currency('RUB')
SEK = Currency('SEK')
SGD = Currency('SGD')
TRY = Currency('TRY')
USD = Currency('USD')
USDT = Currency('USDT')
ZAR = Currency('ZAR')
