from ..types.shape import Shape

#
# Constants
#

arrowdown = Shape('shape_arrow_down')
arrowup = Shape('shape_arrow_up')
circle = Shape('shape_circle')
cross = Shape('shape_cross')
diamond = Shape('shape_diamond')
flag = Shape('shape_flag')
labeldown = Shape('shape_label_down')
labelup = Shape('shape_label_up')
square = Shape('shape_square')
triangledown = Shape('shape_triangle_down')
triangleup = Shape('shape_triangle_up')
xcross = Shape('shape_xcross')
