from ..types.position import Position

#
# Constants
#

bottom_center = Position('bottom_center')
bottom_left = Position('bottom_left')
bottom_right = Position('bottom_right')
middle_center = Position('middle_center')
middle_left = Position('middle_left')
middle_right = Position('middle_right')
top_center = Position('top_center')
top_left = Position('top_left')
top_right = Position('top_right')