from ..types.text import AlignEnum, FormatEnum, WrapEnum

# Text alignment constants
align_left = AlignEnum('left')
align_center = AlignEnum('center')
align_right = AlignEnum('right')
align_top = AlignEnum('top')
align_bottom = AlignEnum('bottom')

# Text format constants
format_bold = FormatEnum('bold')
format_italic = FormatEnum('italic')
format_none = FormatEnum('none')

# Text wrap constants
wrap_auto = WrapEnum('auto')
wrap_none = WrapEnum('none')
